######################################################################
# Copyright (C) 1990
# FourGen Software Technologies, Inc. Edmonds, Washington USA
# Use, modification, duplication, and/or distribution of this
# software is limited to the terms of the software agreement.
# Sccsid:  @(#)  .../user_nul.4gs/l_event.4gl  1.1  Delta: 11/23/90
######################################################################
# Event handling functions.  The functions contained in this file
# work together to perform event and hot key functionality.
# The functions include:
# 
# function event_init()        # Initialize the variables & cursors
# function hot_array()         # Build the hot key array
# function hot_key()           # Process the hot keys
# function hot_local(event)    # Define an event to be processed locally
# function hot_action(act_key) # Perform the action specified by act_key
# function hot_edit()          # Edit hot keys
# function act_pick()          # Pick an action from the picker
# function hot_browse()        # Show current hot key mapping
# function fg_navigate()       # Bootstrap for act_pick()
# function act_edit(act_key)   # Edit an action (navigate command)


globals "globals.4gl"

define
    is_prepared smallint,   # Is SQL statement prepared?
    p_mod char(8),          # Module name
    p_prog char(8),         # Program name
    p_funct char(15),       # Function name
    key_map array[78]       # Hot key map
      of char(15),          #
    keycode array[50]       # List of key codes
      of smallint,          #
    action array[101] of record # List of Actions for Navigate
      act_key char(15),     # action keys
      description char(30)  # descriptions
    end record,
    local_evnt array[20]    # List of events to process locally
      of char(15),          #
    loc_cnt smallint,       # Count of local events in local_evnt array
    event_disable smallint  # (boolean) Disable event processing?

######################################################################
function event_init()
######################################################################
# This function initializes the event handler.  It builds all SQL
# queries, and initializes the event array.
#
    define
        quote char(1),      # storage for the quote character
        num_events smallint,# number of local events
        n smallint          # generic counter

    # Trap fatal errors
    whenever error call error_handler

    # Initialize variables
    let quote = "\""
    let p_mod = ""
    let p_prog = ""

    # Parse the module.program from progid.
    let p_mod = ""
    # Search for the period delimiter
    for n = 1 to 9
        if progid[n] = "."
        then
            let p_mod = progid[1,n-1]
            let p_prog = progid[n+1,17]
            exit for
        end if
    end for
    # Delimiter not found
    if p_prog is null then let p_prog = progid end if

# order    nav_user  nav_program  nav_module
#   1        user       p_prog      p_mod      # user/program specific
#   2        user       null        null       # user global
#   3        null       p_prog      p_mod      # program specific
#   4        null       null        null       # global definition

    # Build the navigation SQL string
    let scratch =
      "select act_key, line_no, nav_user, nav_program, nav_module, ",
      "rowid from stxnvgtd where",
            "   (nav_user = user ",                          # Clause #1
               " and nav_program = ", quote, p_prog clipped, quote,
               " and nav_module = ", quote, p_mod clipped, quote, ") ",
            "or (nav_user = user ",                          # Clause #2
               " and nav_program is null ",
               " and nav_module is null) ",
            "or (nav_user is null ",                         # Clause #3
               " and nav_program = ", quote, p_prog clipped, quote,
               " and nav_module = ", quote, p_mod clipped, quote, ") ",
            "or (nav_user is null ",                         # Clause #4
               " and nav_program is null ",
               " and nav_module is null) ",
        "order by line_no, act_key, nav_user desc, nav_program desc, ",
          "nav_module desc"
    # Prepare the string
    prepare nav_specs from scratch
    declare nav_act cursor for nav_specs

# order    hot_user  hot_program  hot_module
#   1        null       null        null       # global definition
#   2        null       p_prog      p_mod      # program specific
#   3        user       null        null       # user global
#   4        user       p_prog      p_mod      # user/program specific

    # Build the hot key SQL string
    let scratch =
      "select act_key, hot_key, hot_user, hot_program, hot_module ",
      "from stxhotkd where ((hot_key = ?) or (hot_key < ?))",
      "and (   (hot_user = user ",                          # Clause #1
               " and hot_program = ", quote, p_prog clipped, quote,
               " and hot_module = ", quote, p_mod clipped, quote, ") ",
            "or (hot_user = user ",                          # Clause #2
               " and hot_program is null ",
               " and hot_module is null) ",
            "or (hot_user is null ",                         # Clause #3
               " and hot_program = ", quote, p_prog clipped, quote,
               " and hot_module = ", quote, p_mod clipped, quote, ") ",
            "or (hot_user is null ",                         # Clause #4
               " and hot_program is null ",
               " and hot_module is null) ",
            ") ",
        "order by hot_user, hot_program, hot_module"
    # Prepare the string
    prepare find_specs from scratch
    declare hot_curs cursor for find_specs

    # Prepare the action table query
    let scratch = "select description, os_command, press_enter ",
      "from stxactnr where language = \"", get_scrlib("language"),
      "\" and act_key = ?"
    prepare act_specs from scratch
    declare act_curs cursor for act_specs

    # Prepare the key definition table query
    let scratch = "select key_desc, key_code from stxkeysr ",
      "where key_code = ?"
    prepare key_specs from scratch
    declare key_curs cursor for key_specs

    # Build the hot key to event map array
    call hot_array()

    # Default the known local events
    let local_evnt[1] = "tab"
    let local_evnt[2] = "btab"
    let local_evnt[3] = "cancel"
    let local_evnt[4] = "accept"
    let local_evnt[5] = "zoom"
    let local_evnt[6] = "null"
    let loc_cnt = 6

    let event_disable = false  # Don't disable event processing
    let is_prepared = 123
end function
# event_init()

######################################################################
function hot_array()
######################################################################
# This function is called to build the hotkey map array
#
    define
        act_key char(15),   # key into action table
        hot_key smallint,   # key into hot map array
        key_exact smallint, # exact key to read
        key_less  smallint  # read all keys less than this number

    # Read the elements into the key_map[] array
    call wait_on("Loading: Hot Keys")
    let key_exact = 0
    let key_less = 10000
    open hot_curs using key_exact, key_less  # select all keys
    foreach hot_curs into act_key, hot_key
        # Put F Keys into 27-78
        if hot_key > 100 then let hot_key = hot_key - 74 end if
        # Make sure it will fit into the array
        if hot_key > 0 and hot_key < 79
          then let key_map[hot_key] = act_key end if
    end foreach
    call wait_off()

end function
# hot_array()

######################################################################
function hot_key(cur_funct)
######################################################################
# This function is called whenever a control-key or function-key
# is pressed.
# It gets the key that was pressed from the 'hotkey' global variable.
# It's job is to map the key that was just pressed to a function.
# Functions can be local events (zoom,navigate,accept,etc.),
# or O/S type navigation events.
# Most internal events are handled by hot_key(), but some need to
# be handled locally by the specific input/menu/display statement.
# If the event needs to be handled locally by the input/display
# statment (like zoom/tab/cancel/accept), then it sets scr_funct
# and returns (allowing the local function to perform the work).
# The current function is passed as "cur_funct" to this function
# so the program knows where it is.
#
    define
        cur_funct char(15),   # function that called hot_key()
        act_key char(15),     # action key returned from the query
        n     smallint        # generic counter

    # Initialize the event processing arrays and SQL queries
    if is_prepared = 123
      then else call event_init() end if

    # Remember the current function
    let p_funct = cur_funct

    # Work on n, not hotkey
    let n = hotkey

    # Put F Keys into 27-78
    if n > 100 then let n = n - 74 end if

    # Make sure it fits in the array
    if n > 0 and n < 79
      then let act_key = key_map[n]
      else let act_key = "null"
    end if
    
    # Run the event
    call hot_action(act_key)

end function
# hot_key()

######################################################################
function hot_local(event)
######################################################################
# This function is called just before calling hot_key() to inform
# hot_key() to perform the event locally.
# This function tells the hot_action() function to perform "event"
# locally.  (ie: return scr_funct=event vs. running that event if
# the event has been selected)
# You can specify as many as 20 events to process locally.
# You would only specify an event to be local if it would normally
# be processed by hot_action.  No need to specify "zoom", "accept",
# or any other event that is normally processed locally.
#
    define
        event char(15)

    # Initialize the event processing arrays and SQL queries
    if is_prepared = 123
      then else call event_init() end if

    let loc_cnt = loc_cnt + 1
    let local_evnt[loc_cnt] = event
end function
# hot_local()

######################################################################
function hot_action(act_key)
######################################################################
# This function reads the key from the action table, and executes it.
#
    define
        act_key char(15),     # action key returned from the query
        act_cmd char(74),     # action OS command
        act_desc char(30),    # action description
        act_enter char(1),    # press enter after the action? (Y/N)
        n smallint            # generic number

    # Reset scr_funct
    let scr_funct = ""

    # Null event
    if act_key is null then let act_key = "null" end if

    # First, check to see if it is a known local event.
    for n = 1 to loc_cnt
        if act_key = local_evnt[n]
        then
            # Found it.  Process it locally.
            let scr_funct = act_key
            let loc_cnt = 6            # Reset for future calls
            return
        end if
    end for
    let loc_cnt = 6   # Reset for future hot_local() calls

    # (Special case) Don't process the event
    if event_disable = true
    then
        let scr_funct = act_key     # The event that was to be processed
        return
    end if

    # Try to process a known internal event
    if event_prc(act_key,p_funct) = false
    then
        # Try to find the event in stxactnr
        open act_curs using act_key
        fetch act_curs into act_desc, act_cmd, act_enter

        # If it isn't in the table (notfound) or the act_cmd or
        # act_enter columns are null, then it's an internal event.
        if status = notfound or act_enter is null or act_cmd is null
        then
            close act_curs
            # If it's an internal event, then see if it is listed
            # in the global_events() function.
            # If it's not listed there, then return the scr_funct
            # back to the local calling function.
            if global_events(act_key, p_funct)
              then let scr_funct = null      # Already processed
              else let scr_funct = act_key   # Process locally
            end if
            return
        end if
        close act_curs
    
        # Got the action: now run with it.
        let scratch = act_cmd
        if act_enter = "Y"
          then let n = true
          else let n = false
        end if
        call fg_os_exit(act_desc, n)

    end if

    # Reset scr_funct unless "navigate" or "hot browse" was selected
    if act_key = "navigate" or act_key = "hot"
    then else
        let scr_funct = ""
    end if

    return
end function
# hot_action()

######################################################################
function hot_browse()
######################################################################
# This function browses all hot key definitions.  If they choose one,
# it tries to run that event.  If they press [ZOOM], then it allows
# them to edit that hotkey definition.
#
    define
        key_code smallint,  # Key code
        key_desc char(15),  # Key description
        act_desc char(30),  # Action description
        key_ptr smallint,   # Pointer to keycode[] array
        n smallint,         # Generic counter
        m smallint          # Generic counter

    # Initialize the event processing arrays and SQL queries
    if is_prepared = 123
      then else call event_init() end if

    # Lookup the key description
    call wait_on("Loading: Hot Key Definitions")
    let key_ptr = 0

    # Load F-keys first
    for n = 27 to 62
        let key_code = n + 74
        open key_curs using key_code
        fetch key_curs into key_desc
        if status != notfound
        then
            let key_ptr = key_ptr + 1
            let keycode[key_ptr] = key_code
            # Get the event name
            let act_desc = "Undefined"
            open act_curs using key_map[n]
            if status = 0 then fetch act_curs into act_desc end if
            let scratch = key_desc, act_desc
            call textput("(see scratch)")
        end if
    end for

    # Load control keys now
    for key_code = 1 to 26
        open key_curs using key_code
        fetch key_curs into key_desc
        if status != notfound
        then
            let key_ptr = key_ptr + 1
            let keycode[key_ptr] = key_code
            # Get the event name
            let act_desc = "Undefined"
            open act_curs using key_map[key_code]
            if status = 0 then fetch act_curs into act_desc end if
            let scratch = key_desc, act_desc
            call textput("(see scratch)")
        end if
    end for

    # Load special keys (if known)
    for n = 63 to 78
        let key_code = n + 74
        open key_curs using key_code
        fetch key_curs into key_desc
        if status != notfound
        then
            let key_ptr = key_ptr + 1
            let keycode[key_ptr] = key_code
            # Get the event name
            let act_desc = "Undefined"
            open act_curs using key_map[n]
            if status = 0 then fetch act_curs into act_desc end if
            let scratch = key_desc, act_desc
            call textput("(see scratch)")
        end if
    end for

    call wait_off()

    # Call the text picker
    call texthelp("lib_scr","hot_brw",1)
    call textzoom()
    let n = textpick("Hot Keys")

    # Process the picked element
    if n > 0
    then
        # A hot key was selected
        let key_code = keycode[n]
        if key_code > 100
          then let key_code = key_code - 74
        end if
        call hot_action(key_map[key_code])
    else
        # [DEL] or zoom selected.
        if scratch[1,4] = "zoom"
        then
            let n = arr_curr()
            let key_code = keycode[n]
            call hot_edit(key_code)
        end if
    end if

end function
# hot_browse()

######################################################################
function fg_navigate()
######################################################################
# This function allows the user to navigate around different programs.
# It opens a window (picker) to have the user determine where they
# wish to go.  It will also allow the user to customize the menu
# to add the things they most often do.
#
    define
        act_key char(15),  # activation key
        n     smallint     # generic counter

    # Initialize the event processing arrays and SQL queries
    if is_prepared = 123
      then else call event_init() end if

    # Pick the event
    if act_pick()
    then
      let act_key = scratch
      else return
    end if

    # Run the event
    call hot_action(act_key)

end function
# fg_navigate()

function act_edit(act_key)
define act_key char(15), m_act_key char(15), description char(30),
os_command char(74), press_enter char(1), system_wide char(1),
other_access char(1), nav_module char(8), nav_program char(8),
nav_user char(10), is_add smallint, n smallint return end function
function hot_edit(h_key)
define h_key smallint, l_mod char(8), l_prog char(8), l_user char(10),
old_act char(15), key_desc char(15), system_wide char(1),
act_key char(15), act_desc char(30), key_less  smallint,
this_fld char(30), n smallint
call ucl_text("hot_edt") call textview("Hotkey Edit") end function
function act_pick() define act_key char(15), o_act_key char(15),
act_desc char(30), n smallint
call ucl_text("nav_edt") call textview("Navigate")
let scratch = null return false end function
